\ Do not use this file except in compliance with the License. You may
\ obtain a copy of the License at http://www.microcore.org/License/
\ Software distributed under the License is distributed on an "AS IS" basis,
\ WITHOUT WARRANTY OF ANY KIND, either express or implied.
\ See the License for the specific language governing rights and limitations
\ under the License.
\
\ The Original Code is: MicroCore DEBUGGER.F 1.20
\
\ The Initial Developer of the Original Code is Klaus.Schleisiek AT hamburg.de.
\ Extensions by Ulrich.E.Hoffmann AT xlerb.de 
\
\ Introduction  see: doc/uCore_Introduction.pdf
\ Tools         see: doc/uCore_Implementation.pdf
\ Versions      see: uCross_Versions.txt
\ License       see: doc/uCore_License.pdf
\ Conventions   see: doc/uCore_Conventions.pdf
\ Contributors  see: doc/uCore_Contributors.pdf
\
\ MicroCore monitor for the debugger interface
\ hardware must be configured with an appropriate bidirectional debug interface
\

: in?     ( -- f )  status@ #in and ;
: in@     ( -- w )  DEBUG_REG @ ;

: out?    ( -- f )  status@ #out and ;
: out!    ( w -- )  DEBUG_REG ! ;

Variable tmpStatus

: monitor  ( -- )       BEGIN  0 out!  in@ execute  REPEAT ;

: debugger ( -- )      \ main debugger loop called via break instruction
  status@ tmpStatus !
   #smallest out! r> 1- out!  ['] monitor nop BRANCH
;

: cold     ( xx -- )   
   clear
   ['] monitor nop BRANCH
;
: init     ( -- )
   BEGIN  $A0A out!   in@ $5F5 xor 0= UNTIL
   $505 out!  in@ drop \ synchronise Host <-> Target communication
   cold
;
: msg      ( n -- )
    dup out!           \ transfer message number to host
    neg 0=EXIT         \ if negative (error causing abort in host)
    cold               \ then restart debug loop
;
: receivestack ( -- ??? )
   in@ ?dup 0=EXIT       \ number of elements
   1- FOR in@ NEXT
;
: sendstack    ( ??? -- )
   dsp@  -1 #ds_addr_width << invert and
   dup out! ?dup 0=EXIT 1- FOR out!  NEXT
;

: copyrs ( -- ) 
   BEGIN  rsp@ #smallest 1- - WHILE  r> out!  REPEAT
   -1 out!
   BEGIN  in@ dup -1 - WHILE  >r  REPEAT drop
;  
: nextstep ( -- ) r> drop  in@ >r tmpStatus @ status! ;  \ call and jump (modified via >r)
: gettor   ( -- ) r> r>   out! >r ;  \ get tor and transfer to host
: puttor   ( -- ) r> in@    >r >r ;  \ transfer from host and put tor
: copytor  ( -- ) r> r@   out! >r ;  \ get tor and transfer to host

\ opcodes
: \>r          >r          ;
: \r>          r>          ;
: \exit        exit        ;
: \r@          r@          ;
: \r!          r!          ;
: \dup         dup         ;
: \?dup        ?dup        ;
: \drop        drop        ;
: \over        over        ;
: \+           +           ;
: \+c          +c          ;
: \-           -           ;
: \and         and         ;
: \or          or          ;
: \xor         xor         ;
: \nip         nip         ;
: \invert      invert      ;
: \2*          2*          ;
: \2/          2/          ;
: \u2/         u2/         ;
: \ror         ror         ;
: \rol         rol         ;
: \0=          0=          ;
: \+lld        +lld        ;
: \+lst        +lst        ;
: \status!     status!     ;
: \status@     status@     ;
: \ip!         ip!         ;
: \ip@         ip@         ;
: \+tst        +tst        ;
: \+tld        +tld        ;
: \dsp!        dsp!        ;
: \dsp@        dsp@        ;
: \rsp!        rsp!        ;
: \rsp@        rsp@        ;
\ macros
: \under       under       ;
: \nor@        nor@        ;                 
: \@           @           ;                 
: \!           !           ;                 
: \l@          l@          ;
: \l!          l!          ;
: \t@          t@          ;
: \t!          t!          ;
: \ie!         ie!         ;
: \flags@      flags@      ;
: \task@       task@       ;
: \task!       task!       ;
: \rdrop       rdrop       ;
: \2drop       2drop       ;
: \2dup        2dup        ;
: \c=0         c=0         ;
: \c=1         c=1         ;
: \=           =           ;
: \negate      negate      ;
: \tuck        tuck        ; 
: \swap        swap        ;
: \rot         rot         ;
: \-rot        -rot        ;
