\ Do not use this file except in compliance with the License. You may
\ obtain a copy of the License at http://www.microcore.org/MPL/
\ Software distributed under the License is distributed on an "AS IS" basis,
\ WITHOUT WARRANTY OF ANY KIND, either express or implied.
\ See the License for the specific language governing rights and limitations
\ under the License.
\
\ The Original Code is: MicroCore DISASM.F 1.20
\
\ The Initial Developer of the Original Code is Klaus.Schleisiek AT hamburg.de.
\ Port to the gforth system and extensions by Ulrich.E.Hoffmann AT xlerb.de
\
\ Introduction  see: doc/uCore_Introduction.pdf
\ Tools         see: doc/uCore_Implementation.pdf
\ Versions      see: uCross_Versions.txt
\ License       see: doc/uCore_License.pdf
\ Conventions   see: doc/uCore_Conventions.pdf
\ Contributors  see: doc/uCore_Contributors.pdf
\
\ MicroCore Disassembler with Macro resolution
\ DISASM displays one instruction per key pressed. <cr> or <esc> to finish
\ This code loads on the public domain Win32Forth http://www.forth.org/
\

Only Forth also definitions decimal

Variable Expand  Expand off  \ OFF: display macros with macro name. 
                             \ ON: display every single instruction
Variable out     out off

: dis_cr     ( -- )            out off    _CR ;
: dis_emit   ( char -- )       1 out +!   _EMIT ;
: dis_type   ( addr len -- )   dup out +! _TYPE ;
: position   ( n -- )
   out @ max BEGIN  dup out @ - WHILE  BL dis_emit  REPEAT  drop ;

: results    ( -- )   018 position ;

: std_output ( -- )  [']    _CR IS CR  [']    _EMIT IS EMIT  [']    _TYPE IS TYPE ;
: dis_output ( -- )  ['] dis_CR IS CR  ['] dis_EMIT IS EMIT  ['] dis_TYPE IS TYPE ;

: .stack  ( opcode -- )  #stack and
   #none case? IF  ."  NONE"  EXIT THEN
   #push case? IF  ."  PUSH"  EXIT THEN
   #pop  =     IF  ."  POP "  EXIT THEN  ."  BOTH" ;

: alucode  ( opcode -- opcode )  dup #group and
   over #stack and #both =                 \ unary operators
   IF   #not  case? IF  ." NOT   "  EXIT THEN
        #sl   case? IF  ." SL    "  EXIT THEN
        #asr  case? IF  ." ASR   "  EXIT THEN
        #lsr  case? IF  ." LSR   "  EXIT THEN
        #ror  case? IF  ." ROR   "  EXIT THEN
        #rol  case? IF  ." ROL   "  EXIT THEN
        #zequ     = IF  ." ZEQU  "  EXIT THEN
                        ." CC    "  EXIT
   THEN
   #add  case? IF  ." ADD   "  EXIT THEN
   #sub  case? IF  ." SUB   "  EXIT THEN
   #adc  case? IF  ." ADC   "  EXIT THEN
   #ssub case? IF  ." SSUB  "  EXIT THEN
   #and  case? IF  ." AND   "  EXIT THEN
   #or   case? IF  ." OR    "  EXIT THEN
   #xor      = IF  ." XOR   "  EXIT THEN
   ( #nos )       ." NOS   "
;
: bracode ( opcode -- opcode )   dup #group and
   over #stack and #push =                       \ branch operators
   IF  #dup     case? IF  ." DUP   "  EXIT THEN
       #exc     case? IF  ." EXC   "  EXIT THEN
       #?dup    case? IF  ." ?DUP  "  EXIT THEN
       #?ovl    case? IF  ." ?OVL  "  EXIT THEN
       #int     case? IF  ." INT   "  EXIT THEN
       #iret    case? IF  ." IRET  "  EXIT THEN
       #thread      = IF  ." THREAD"  EXIT THEN
                          ." TOKEN "  EXIT
   THEN                                          \ conditions
   #never   case? IF  ." NEVER "  EXIT THEN
   #zero    case? IF  ." ZERO  "  EXIT THEN
   #nzero   case? IF  ." NZERO "  EXIT THEN
   #sign    case? IF  ." SIGN  "  EXIT THEN
   #nsign   case? IF  ." NSIGN "  EXIT THEN
   #novl    case? IF  ." NOVL  "  EXIT THEN
   #ncarry      = IF  ." NCARRY"  EXIT THEN
   ( #always )        ." ALWAYS"
;
: memcode  ( opcode -- opcode )  dup #group and
  #status   case? IF  ." STATUS"  EXIT THEN
  #tor      case? IF  ." TOR   "  EXIT THEN
  #rstack   case? IF  ." RSTACK"  EXIT THEN
  #local    case? IF  ." LOCAL "  EXIT THEN
  #rsp      case? IF  ." RSP   "  EXIT THEN
  #dsp      case? IF  ." DSP   "  EXIT THEN
  #task         = IF  ." TASK  "  EXIT THEN
                      ." IP    "
;
: .colon  ( caddr -- caddr | ff )
   >r  Colons
   BEGIN  @ ?dup
   WHILE  dup cell- @ r@ =
          IF  cell- body> >name .name  r> 0=  EXIT THEN
   REPEAT  r> ;

: call_inst?  ( caddr n -- caddr n ff | caddr+1 tf )
   over opcode@ [ Asm ALWAYS BOTH BRA H ] Literal = dup 0=EXIT
   [ Asm ALWAYS BOTH BRA H ] Literal .opcode  results
   swap >r  swap 1+ tuck r> +
   .colon ?dup 0=EXIT  .addr ."  CALL" ;

: .brnop  ( caddr n opcode -- caddr+1 tf )
   .opcode  results  over + 1+ .addr space  1+ True ;

: conditions?  ( caddr n -- caddr n ff | caddr+1 tf )
   over opcode@
   [ Asm ALWAYS POP BRA H ] Literal over = IF  .brnop ." BRANCH"    EXIT THEN
   [ Asm NEVER  POP BRA H ] Literal over = IF  .opcode swap 1+ swap over
                                               opcode@ dup >r .brnop
                                               r> 1 and IF  ." 0<>BRANCH" EXIT THEN
                                                            ." 0=BRANCH"  EXIT
                                           THEN
   [ Asm ZERO   POP BRA H ] Literal over = IF  .brnop ." Z-BRANCH"  EXIT THEN
WITH_TOR [IF]
   [ Asm NZERO  POP BRA H ] Literal over = IF  .brnop ." NEXT" EXIT THEN
[ELSE]
   [ Asm NZERO  POP BRA H ] Literal over = IF  .brnop ." NZ-BRANCH" EXIT THEN
[THEN]
   [ Asm SIGN   POP BRA H ] Literal over = IF  .brnop ." S-BRANCH"  EXIT THEN
   [ Asm NSIGN  POP BRA H ] Literal over = IF  .brnop ." NS-BRANCH" EXIT THEN
   [ Asm NOVL   POP BRA H ] Literal over = IF  .brnop ." NO-BRANCH" EXIT THEN
   [ Asm NCARRY POP BRA H ] Literal over = IF  .brnop ." NC-BRANCH" EXIT THEN
   drop False
;
: .literal  ( caddr -- caddr' )
   dup nibbles@ 0 ?DO  .opcode  LOOP
   dup >branch drop
   swap nibbles@ nibbles>
   call_inst?    ?EXIT
   conditions?  ?EXIT
   over opcode@ nop? IF  [ Asm NEVER NONE BRA H ] Literal .opcode  swap 1+ swap  THEN
   results dup .addr space .
;
: operator?  ( opcode -- f )
   Operators
   BEGIN  @ ?dup
   WHILE  2dup 2 cells - @ =
          IF  2 cells -  body> >name results .name  nip  True  EXIT THEN
   REPEAT  drop False ;

: macro?  ( caddr -- caddr' tf | caddr ff )
   Expand @ IF  False  EXIT THEN
   dup macro@ dup 0=EXIT  >r
   BEGIN  dup macro@ r@ = WHILE  dup opcode@ .opcode   1+  REPEAT
   results   Macros
   BEGIN  @ ?dup
   WHILE  dup cell- @ r@ =
          IF  2 cells -  body> >name  .name  r> EXIT THEN
   REPEAT ." macro"  r> dup .
;
: .autoinc  ( n1 -- )  
   #group and dup 4 and
   IF  -1 2* 2* or  ELSE  space  THEN  . 
;
: .instruction  ( caddr -- caddr' )
   cr dup .colon 0= IF  results ." ---------------" cr  THEN
   dup .addr ." : "
   macro? ?EXIT
   dup opcode@ #literal and IF  .literal  EXIT THEN
   dup 1+ swap opcode@   dup .opcode
   dup operator? IF  drop  EXIT THEN
   results   dup #op and
   #alu case? IF  alucode .stack ."  ALU" EXIT THEN
   #bra case? IF  bracode .stack ."  BRA" EXIT THEN
   #mem case? IF  dup #stack and #push case? IF  .autoinc ." +ld"  EXIT THEN
                                 #pop  =     IF  .autoinc ." +st"  EXIT THEN
                  memcode .stack ."  MEM" EXIT
              THEN
   drop       ." USR " #usrmask and . ;

: break_key?  ( -- f )   key dup #cr = swap #esc = or ;
\ : stop?       ( -- f )   key? dup 0=EXIT  drop  break_key? dup ?EXIT  drop break_key? ;

: disasm  ( caddr -- )  \ displays one instruction per key pressed. <cr> or <esc> to finish
   Base @ swap  hex  dis_output
   BEGIN  .instruction  break_key? UNTIL
   drop  Base !  std_output 
;
: show    ( <name> -- )  Target defined Host 0= ?missing >body @ disasm
;
: decode  ( opcode -- )  dup #op and
   #bra case? IF  bracode .stack ."  BRA" EXIT THEN
   #alu case? IF  alucode .stack ."  ALU" EXIT THEN
   #mem     = IF  memcode .stack ."  MEM" EXIT THEN
   #usrmask and trap-vector u.   ."  USR"
;

